#!/usr/bin/tclsh

# Brickshelf Backup Crawler 1.1
# 17 July 2007 by Jim DeVona
# A utility to help batch download images and files from Brickshelf.
# Presumably similar in function to "BSBackup":
# http://www.fial.com/bob/partscatalog/download.php
# 1.1 bug fix: adds support for multi-page m=username galleries (duh)

package require http
package require struct
package require htmlparse

# support for https per http://wiki.tcl.tk/2627
# package require tls
# ::http::register https 448 ::tls::socket

#
# Download
#
# Parameters:
#	url, address of file to download
#	dir, local directory to save downloaded file
#	name, local name for downloaded file
#
# Return:
#	nothing if ok
#	otherwise error
#
proc Download { url dir name } {
	global pause
	
	# validate download location
	if {![file isdirectory $dir]} {
		return -code error "$dir is not a directory"
	}
	
	set path [file join $dir $name]
	# avoid overwriting existing files
	if {[file exists $path]} {
		puts "$path already exists"
		return
	}
		
	# download file
	puts "Downloading $url to $path"
	if {[catch {open $path w} out]} {
		return -code error $out
	}
	if {[catch {::http::geturl $url -channel $out} token]} {
		return -code error $token
	}
	close $out
	::http::cleanup $token

	# throttle
	after $pause
}

#
# ParsePage
#
# Parameters:
#	mode, m or f (user or folder gallery)
#	curfolder, id of current brickshelf folder
#	curdir, path to corresponding local directory
#	start, list folder starting which what item
#		(optional; used to traverse multi-page galleries)
#
# Return:
#	nothing if ok
#	otherwise error
#
proc ParsePage { mode curfolder curdir {start 0} } {
	
	# validate local directory
	if {![file isdirectory $curdir]} {
		return -code error "$curdir is not a directory"
	}
	
	# what page are we about to retrieve?
	set pageurl "http://www.brickshelf.com/cgi-bin/gallery.cgi?$mode=$curfolder"
	set status "Processing folder $curfolder"
	if {$start > 0} {
		append pageurl "&n=$start"
		append status " (page beginning after $start items)"
	}
	puts $status
	
	# retrieve the page
	if {[catch {::http::geturl $pageurl} token]} {
		return -code error $token
	}
	
	# parse the page and release the initial query
	set doc [::struct::tree]
	if {[catch {::htmlparse::2tree [::http::data $token] $doc} err]} {
		return -code error $err
	}
	::http::cleanup $token
		
	# process page contents
 	$doc walk root {node} {
 		set type [$doc get $node "type"]
 		if {$type eq "a"} {
 			set data [$doc get $node "data"]
 			set location [GetAttr $data "href"]
 			if {[regexp {^/cgi-bin/gallery\.cgi\?f=(\d+)$} $location {} subfolder]} {
 				# link to subfolder
 				foreach child [$doc children $node] {
 					if {[$doc get $child "type"] eq "img"} {
 						set imginfo [$doc get $child "data"]
 						set alt [GetAttr $imginfo "alt"]
 						set subdir [file join $curdir $alt]
 						if {![file isdirectory $subdir]} {
 							file mkdir $subdir
 						}
 						if {[catch {ParsePage "f" $subfolder $subdir} err]} {
 							return -code error $err
 						}
 					}
 				}
 			} elseif {[regexp {^/cgi-bin/gallery\.cgi\?([mf])=([a-zA-Z0-9\-_@\.]+)&n=(\d+)$} $location {} nextmode nextfolder next]} {
 				# link to next page of this folder
 				if {($nextfolder eq $curfolder) && ($nextmode eq $mode)} {
 					# only follow actual "next" links, not "prev" links
	 				if {$next > $start} {
	 					# continue processing contents of next page in same local directory
		 				if {[catch {ParsePage $mode $curfolder $curdir $next} err]} {
		 					return -code error $err
		 				}
	 				}
	 			} else {
	 				# linking to a different page of a different folder
	 				# doesn't usually seem to happen, but we ralph if so
	 				return -code error "Cross link to $nextfolder item $next"
	 			}
 			} elseif {[regexp {^/cgi-bin/gallery\.cgi\?i=\d+$} $location {}] ||
 					[regexp {^/gallery} $location {}]} {
 				# image or file download
 				foreach child [$doc children $node] {
 					if {[$doc get $child "type"] eq "img"} {
 						set imginfo [$doc get $child "data"]
 						set alt [GetAttr $imginfo "alt"]
 						set src [GetAttr $imginfo "src"]
 						set path [GetPath $src]
 						set url [format "http://www.brickshelf.com%s" $path]
 						if {[catch {Download $url $curdir $alt} err]} {
 							# proceed after failed downloads
 							puts $err
 						}
 					}
 				}
 			}
 		}
 	}
	
	$doc destroy
}

# Stupid little utility to return the value of an attribute from
# a chunk of tag data (which htmlparse does not parse, phooey).
# Split key=value pairs on whitespace; ASSUMES NO WHITESPACE IN
# VALUE (because Brickshelf doesn't use any in the relevant cases).
# Trims quotes around value, if any, and returns value, or "".
proc GetAttr { data key } {
	set attrs [split $data]
	foreach attr $attrs {
		if {[regexp -nocase "^$key=\"?(\[^\"]*)\"?$" $attr {} value]} {
			return $value
		}
	}
	return ""
}

# Stupid little utility to convert the path to a file thumbnail
# to the path to file itself: excise "/thumb" subdirectory and
# trim trailing "_thumb.jpg" suffix. Returns file path.
proc GetPath { src } {
	set f [string first "/thumb" $src]
	return [string range [string replace $src $f [expr {$f+5}]] 0 end-10]
}	

# start

# usage
if {$argc == 0} {
	set name [file tail $argv0]
	puts stderr "Usage: $name (-m user|-f folder|-url URL) -dir directory \[-pause seconds\]"
	exit 0
}

# initialize
set mode {}
set folder {}
set directory {}
set pause 1500

# read parameters
for {set argi 0} {$argi < $argc} {incr argi} {
	switch -- [lindex $argv $argi] {
		-m {
			if {$argi + 1 >= $argc} {
				puts stderr "Missing username argument for -m"
				exit 1
			}
			set mode "m"
			set folder [lindex $argv [incr argi]]
		}
		-f {
			if {$argi + 1 >= $argc} {
				puts stderr "Missing folder id argument for -f"
				exit 1
			}
			set mode "f"
			set folder [lindex $argv [incr argi]]
		}
		-url {
			if {$argi + 1 >= $argc} {
				puts stderr "Missing URL argument for -url"
				exit 1
			}
			set url [lindex $argv [incr argi]]
			if {![regexp -nocase {^http://www\.brickshelf\.com/cgi-bin/gallery\.cgi\?([mf])=([a-zA-Z0-9\-_@\.]+)$} $url {} mode folder]} {
				puts stderr "Invalid URL: $url"
				exit 1
			}
		}
		-dir {
			if {$argi + 1 >= $argc} {
				puts stderr "Missing directory argument for -dir"
				exit 1
			}
			set directory [lindex $argv [incr argi]]
		}
		-pause {
			if {$argi + 1 >= $argc} {
				puts stderr "Missing seconds argument for -pause"
				exit 1
			}
			scan [lindex $argv [incr argi]] "%f" seconds
			if {$seconds > 0} {
				set pause [expr {int($seconds * 1000)}]
			}
		}
		default {
			puts stderr "Unrecognized option: [lindex $argv $argi]"
			exit 1
		}
	}
}

# confirm that we got all necessary parameters
if {$mode == {} || $folder == {}} {
	puts stderr "No gallery specified. Use -m, -f, or -url."
	exit 1
}
if {$directory == {}} {
	puts stderr "No local directory specified. Use -dir."
	exit 1
}

# begin crawling
if {[catch {ParsePage $mode $folder $directory} err]} {
	puts stderr "Error: $err"
	exit 1
}
